@@ -1,5 +1,5 @@
 /*
- * Copyright 2005-2021 Gentoo Foundation
+ * Copyright 2005-2022 Gentoo Foundation
  * Distributed under the terms of the GNU General Public License v2
  *
  * Copyright 2005-2010 Ned Ludd        - <solar@gentoo.org>
@@ -126,18 +126,18 @@ parse_date(const char *sdate, time_t *t)
 		 */
 		size_t len = strspn(sdate, "0123456789-:T@");
 		if (sdate[len] == '\0') {
-			const char *fmt;
 			if (sdate[0] == '@') {
-				fmt = "@%s";
+				time_t d = (time_t)strtoll(&sdate[1], (char **)&s, 10);
+				localtime_r(&d, &tm);
 			} else if (strchr(sdate, '-') == NULL) {
-				fmt = "%s";
+				time_t d = (time_t)strtoll(sdate, (char **)&s, 10);
+				localtime_r(&d, &tm);
 			} else if ((s = strchr(sdate, 'T')) == NULL) {
-				fmt = "%Y-%m-%d";
+				s = strptime(sdate, "%Y-%m-%d", &tm);
 			} else {
-				fmt = "%Y-%m-%dT%H:%M:%S";
+				s = strptime(sdate, "%Y-%m-%dT%H:%M:%S", &tm);
 			}
 
-			s = strptime(sdate, fmt, &tm);
 			if (s == NULL || s[0] != '\0')
 				return false;
 		} else {
@@ -797,7 +797,9 @@ static int do_emerge_log(
 							|| flags->do_running)
 					{
 						/* find in list of averages */
-						if (flags->do_predict || verbose) {
+						if (flags->do_predict ||
+							(verbose && !flags->do_running))
+						{
 							snprintf(afmt, sizeof(afmt), "%s/%s",
 									pkgw->atom->CATEGORY, pkgw->atom->PF);
 						} else {
@@ -940,7 +942,9 @@ static int do_emerge_log(
 							|| flags->do_running)
 					{
 						/* find in list of averages */
-						if (flags->do_predict || verbose) {
+						if (flags->do_predict ||
+							(verbose && !flags->do_running))
+						{
 							snprintf(afmt, sizeof(afmt), "%s/%s",
 									pkgw->atom->CATEGORY, pkgw->atom->PF);
 						} else {
@@ -1255,7 +1259,7 @@ static int do_emerge_log(
 								continue;
 							}
 						case NEWER:
-							if (atom->pfx_op != ATOM_OP_NEWER ||
+							if (atom->pfx_op != ATOM_OP_NEWER &&
 									atom->pfx_op != ATOM_OP_NEWER_EQUAL)
 								continue;
 							/* fall through */
@@ -1676,7 +1680,7 @@ int qlop_main(int argc, char **argv)
 		m.do_unmerge = 1;
 		if (array_cnt(atoms) == 0) {
 			m.do_sync = 1;
-			if (start_time == 0)
+			if (start_time == -1)
 				m.show_lastmerge = 1;
 		}
 		m.do_time = 1;
