@@ -1,5 +1,5 @@
 /*
- * Copyright 2005-2021 Gentoo Authors
+ * Copyright 2005-2022 Gentoo Authors
  * Distributed under the terms of the GNU General Public License v2
  *
  * Copyright 2005-2010 Ned Ludd        - <solar@gentoo.org>
@@ -129,45 +129,32 @@ fetch(const char *destdir, const char *src)
 		char buf[BUFSIZ];
 		snprintf(buf, sizeof(buf), "(export DISTDIR='%s' URI='%s/%s'; %s)",
 			destdir, binhost, src, getenv("FETCHCOMMAND"));
-		xsystem(buf);
+		xsystem(buf, AT_FDCWD);
 	} else
 #endif
 	{
-		pid_t p;
-		int status;
-		char *path;
-
-		xasprintf(&path, "%s/%s", binhost, src);
+		char *path = NULL;
 
 		/* wget -c -q -P <dir> <uri> */
-		char prog[] = "wget";
-		char *const argv[] = {
-			prog,
-			(char *)"-c",
-			(char *)"-P",
-			(char *)destdir,
+		const char *argv[] = {
+			"echo",
+			"wget",
+			"-c",
+			"-P",
+			destdir,
 			path,
 			quiet ? (char *)"-q" : NULL,
 			NULL,
 		};
-		if (!(force_download || install) && pretend)
-			strcpy(prog, "echo");
 
-		p = vfork();
-		switch (p) {
-		case 0:
-			_exit(execvp(prog, argv));
-		case -1:
-			errp("vfork failed");
-		}
+		xasprintf(&path, "%s/%s", binhost, src);
+
+		if (!pretend && (force_download || install))
+			xsystemv(&argv[1], AT_FDCWD);  /* skip echo */
+		else
+			xsystemv(argv, AT_FDCWD);
 
 		free(path);
-
-		waitpid(p, &status, 0);
-#if 0
-		if (WIFEXITED(status) && WEXITSTATUS(status) == 0)
-			return;
-#endif
 	}
 
 	fflush(stdout);
@@ -767,7 +754,7 @@ pkg_run_func_at(
 		/*7*/ phase_replacingvers[phaseidx].varname,
 		/*8*/ replacing,
 		/*9*/ debug ? "set -x;" : "");
-	xsystembash(script, dirfd);
+	xsystem(script, dirfd);
 	free(script);
 }
 #define pkg_run_func(...) pkg_run_func_at(AT_FDCWD, __VA_ARGS__)
@@ -1034,6 +1021,7 @@ pkg_merge(int level, const depend_atom *qatom, const tree_match_ctx *mpkg)
 				ATOM_COMP_NOSUBSLOT | ATOM_COMP_NOREPO);
 		replver   = previnst->atom->PVR;
 	}
+	atom_implode(slotatom);
 
 	(void)qprint_tree_node(level, mpkg, previnst, replacing);
 
@@ -1093,11 +1081,18 @@ pkg_merge(int level, const depend_atom *qatom, const tree_match_ctx *mpkg)
 		freeargv(ARGC, ARGV);
 	}
 
+	if (pretend == 100) {
+		tree_match_close(previnst);
+		return;
+	}
+
 	/* Get a handle on the main vdb repo */
 	vdb = tree_open_vdb(portroot, portvdb);
 	if (vdb == NULL) {
-		if (pretend)
+		if (pretend) {
+			tree_match_close(previnst);
 			return;
+		}
 		/* try to create a vdb if none exists yet */
 		xasprintf(&p, "%s/%s", portroot, portvdb);
 		mkdir_p(p, 0755);
@@ -1110,12 +1105,14 @@ pkg_merge(int level, const depend_atom *qatom, const tree_match_ctx *mpkg)
 	if (!cat_ctx) {
 		if (errno != ENOENT) {
 			tree_close(vdb);
+			tree_match_close(previnst);
 			return;
 		}
 		mkdirat(vdb->tree_fd, mpkg->atom->CATEGORY, 0755);
 		cat_ctx = tree_open_cat(vdb, mpkg->atom->CATEGORY);
 		if (!cat_ctx) {
 			tree_close(vdb);
+			tree_match_close(previnst);
 			return;
 		}
 	}
@@ -1143,6 +1140,7 @@ pkg_merge(int level, const depend_atom *qatom, const tree_match_ctx *mpkg)
 		if (vdbfd == -1)
 			err("failed to open vdb extraction directory");
 		tbz2size = xpak_extract(mpkg->path, &vdbfd, pkg_extract_xpak_cb);
+		close(vdbfd);
 	}
 	if (tbz2size <= 0)
 		err("%s appears not to be a valid tbz2 file", mpkg->path);
@@ -1517,9 +1515,6 @@ pkg_unmerge(tree_pkg_ctx *pkg_ctx, depend_atom *rpkg, set *keep,
 	printf("%s***%s unmerging %s\n", YELLOW, NORM,
 			atom_format("%[CATEGORY]%[PF]", tree_get_atom(pkg_ctx, false)));
 
-	if (pretend == 100)
-		return 0;
-
 	portroot_fd = cat_ctx->ctx->portroot_fd;
 
 	/* execute the pkg_prerm step if we're just unmerging, not when
@@ -1527,12 +1522,13 @@ pkg_unmerge(tree_pkg_ctx *pkg_ctx, depend_atom *rpkg, set *keep,
 	 * the replacement package */
 	if (!pretend && rpkg == NULL) {
 		buf = tree_pkg_meta_get(pkg_ctx, EAPI);
+		if (buf == NULL)
+			buf = (char *)"0";  /* default */
 		phases = tree_pkg_meta_get(pkg_ctx, DEFINED_PHASES);
-		buf = tree_pkg_meta_get(pkg_ctx, EAPI);  /* when phases caused ralloc */
 		if (phases != NULL) {
 			mkdirat(pkg_ctx->fd, "temp", 0755);
 			pkg_run_func_at(pkg_ctx->fd, ".", phases, PKG_PRERM,
-					T, T, buf, rpkg == NULL ? "" : rpkg->PVR);
+							T, T, buf, "");
 		}
 	}
 
@@ -1667,9 +1663,13 @@ pkg_unmerge(tree_pkg_ctx *pkg_ctx, depend_atom *rpkg, set *keep,
 	if (!pretend) {
 		buf = tree_pkg_meta_get(pkg_ctx, EAPI);
 		phases = tree_pkg_meta_get(pkg_ctx, DEFINED_PHASES);
-		/* execute the pkg_postrm step */
-		pkg_run_func_at(pkg_ctx->fd, ".", phases, PKG_POSTRM,
-				T, T, buf, rpkg == NULL ? "" : rpkg->PVR);
+		if (buf == NULL)
+			buf = (char *)"0";  /* default */
+		if (phases != NULL) {
+			/* execute the pkg_postrm step */
+			pkg_run_func_at(pkg_ctx->fd, ".", phases, PKG_POSTRM,
+					T, T, buf, rpkg == NULL ? "" : rpkg->PVR);
+		}
 
 		/* finally delete the vdb entry */
 		rm_rf_at(pkg_ctx->fd, ".");
@@ -1693,8 +1693,9 @@ unlink_empty(const char *buf)
 	if (fd != -1 && stat(buf, &st) != -1) {
 		if (st.st_size == 0)
 			ret = unlink(buf);
-		close(fd);
 	}
+	if (fd != -1)
+		close(fd);
 	return ret;
 }
 
@@ -1769,7 +1770,6 @@ pkg_fetch(int level, const depend_atom *qatom, const tree_match_ctx *mpkg)
 	if (pretend) {
 		if (!install)
 			install++;
-		/* qprint_tree_node(level, qatom, mpkg); */
 		pkg_merge(level, qatom, mpkg);
 		return;
 	}
@@ -2054,7 +2054,7 @@ int qmerge_main(int argc, char **argv)
 		int save_verbose = verbose;
 		int save_quiet = quiet;
 
-		pretend = 100;
+		pretend = save_pretend ? 10 : 100;
 		verbose = 0;
 		quiet = 1;
 		ret = qmerge_run(todo);
@@ -2077,5 +2077,11 @@ int qmerge_main(int argc, char **argv)
 	ret = qmerge_run(todo);
 	if (todo != NULL)
 		free_set(todo);
+
+	if (_qmerge_binpkg_tree != NULL)
+		tree_close(_qmerge_binpkg_tree);
+	if (_qmerge_vdb_tree != NULL)
+		tree_close(_qmerge_vdb_tree);
+
 	return ret;
 }
